
import matplotlib.pyplot as plt
import numpy as np
import scipy.integrate
import scipy.ndimage as ndi

import doas
from . import misc
from . import plume


import scipy.optimize as sopt




def distance(q,p):
    return np.sqrt((q[0] - p[0])**2 + (q[1] - p[1])**2)


def find_min_distance_to_curve(p, f, c, x0, y0):
    """
    p :: point x0,y0
    f :: function
    c :: coefficients of 2D curve
    """
    constr = [{
        'type': 'eq',
        'fun':  lambda q,c: f(c, x=q[0], y=q[1], x0=x0, y0=y0)[0] - q[0],
        'args': (c,)
    },{
        'type': 'eq',
        'fun':  lambda q,c: f(c, x=q[0], y=q[1], x0=x0, y0=y0)[1] - q[1],
        'args': (c,)
    }]

    q0 = f(c, x=p[0], y=p[1], x0=x0, y0=y0)

    res = sopt.minimize(distance, x0=q0, args=(p,),
            method='SLSQP', constraints=constr[0],
            options={'maxiter': 100},
    )

    if not res.success:
        print(res)

    return res.x, res.fun, res



def gauss(p, x,y,u, x0=0.0, y0=0.0):
    """\
        Computes plume in kg/m2.

        x: distance from plume (in km)
        y: distance from center line (in km)

        u: wind speed (in meters)
    """
    # conversion factor (kg/m2 -> ppm)   
    air_column = 1.65e29 # molecules per m2
    m = 44.01e-3         # CO2 molar mass: kg/mol
    NA = 6.02214129e23   # 

    c = 1e6 / m / (air_column / NA)

    # km -> m
    x = x * 1e3 # km -> m
    y = y * 1e3 # km -> m

    # state vector
    Q = p[0] / 0.031536  # emissions in Mt/y (convert to kg/s)
    K = p[1]             # eddy diffusion coefficient (in m^2 / s)
    BG = p[2]            # background field (in ppm)

    # dispersion along the plume
    sigma = np.sqrt(2.0 * K * (x-x0) / u)

    # compute Gaussian plume
    A = Q / (np.sqrt(2.0 * np.pi) * sigma * u)
    B = np.exp(-0.5 * (y-y0)**2 / sigma**2)

    p = A * B
    p[x <= 0.0] = 0.0

    # kg/m2 -> ppm
    p = c*p + BG

    return p



def find_all_roots(hh, dd, d, h_max=5000.0):
    """\
    Find all heights (hh), below hh_max where wind direction (dd) equals
    given wind direction (d).
    """

    founds = []

    for i in range(hh.size-1):
        if dd[i] < dd[i+1]:
            dmin, dmax = dd[i], dd[i+1]
            hmin, hmax = hh[i], hh[i+1]
        else:
            dmin, dmax = dd[i+1], dd[i]
            hmin, hmax = hh[i+1], hh[i]

        if hmin > h_max:
            continue

        if dmin <= d < dmax:
            d0 = np.interp(d, [dmin, dmax], [hmin, hmax])
            founds.append(d0)

    return founds




def arc_length_of_poly(f, c, a, b):

    x = np.linspace(a,b)

    s = np.sqrt( 1 + np.polyval(np.polyder(c), x)**2 )

    return scipy.integrate.simps(s,x)


def arc_length_of_2d_curve(f, c, a, b):

    t = np.linspace(a,b, 5001)
    xt, yt = f(c,t, m=1)

    v = np.sqrt( xt**2 + yt**2)

    l = scipy.integrate.simps(v,t)
    return l




def compute_arc_length_and_distance(points, func, coeff,
        x_o=0.0, y_o=0.0, x0=0.0, y0=0.0):
    """\
    Given n points and polynomial function on [x_o, x_max], compute
    arc length along curve and distance from curve.
    """
    dxs = []
    dys = []

    t0 = np.sqrt((x_o - x0)**2 + (y_o - y0)**2)

    for px,py in points:
        (qx,qy), dy, res = find_min_distance_to_curve([px,py], func, coeff,
                x0, y0)

        t = np.sqrt((qx - x0)**2 + (qy - y0)**2)
        dx = arc_length_of_2d_curve(func, coeff, t0, t)

        dxs.append(dx)
        dys.append(dy)

    return np.array(dxs), np.array(dys)





def flux_inversion(p0, dx, dy, field, wind, std=0.7):

    def model(x,b):
        return gauss(x,*b)

    p0 = np.asarray(p0)
    Se = np.full(field.shape, std**2)

    b = (dx,dy,wind)
    x, res = doas.solver.gauss_newton(p0, field, model, Se=Se, b=b)

    return x, res




def poly2d(c, t=None, x=None, y=None, x0=0.0, y0=0.0, m=0):

    # TODO: estimate x0 and y0 from data

    if t is None:
        t = np.sqrt((x-x0)**2 + (y-y0)**2)
        #t = np.arctan2(y - 30, x - 10)

    degree = c.size // 2
    cx = c[:degree]
    cy = c[degree:]

    if m != 0:
        cx = np.polyder(cx, m)
        cy = np.polyder(cy, m)

    xt = np.polyval(cx,t)
    yt = np.polyval(cy,t)

    return xt, yt



def fit_2D_curve(x,y,w, degree=2, x0=0.0, y0=0.0):

    def objective(c, t, w, x, y, x0, y0):
        xt,yt = poly2d(c,x=x,y=y, x0=x0, y0=y0)
        return np.concatenate([w*(xt-x), w*(yt-y)])


    # add origin
    x = np.append(x, x0)
    y = np.append(y, y0)
    w = np.append(w, 1000.)

    # distance from origin
    r = np.sqrt((x - x0)**2 + (y - y0)**2)

    # angle around origin (not used)
    phi = np.arctan2(y - y0, x - x0)

    # curve fit
    c0 = np.zeros(2 * (degree + 1))
    c, res = sopt.leastsq(objective, x0=c0, args=(r,w,x,y,x0,y0))

    if res not in [1,2,3,4]:
        print('least square failed with error code: ', res)

    # compute function: x(t), y(t)
    #t = np.linspace(r.min(), r.max())
    xt, yt = poly2d(c, x=x, y=y, x0=x0, y0=y0)

    return xt, yt, poly2d, c




def transform_to_kilometers(lon, lat, lon_o, lat_o, epsg=3006):
    """\
    Transform WGS84 to meters (using EPSG system).
    """
    x_o, y_o = misc.transform_points(lon_o, lat_o, epsg)
    x, y = misc.transform_points(lon, lat, epsg)
    return (x - x_o) / 1e3, (y - y_o) / 1e3


def setup_flux_inversion(lon0, lat0, field, clouds, is_plume, t_values, noise,
        lon_model, lat_model, height_model, uwind, vwind,
        origin=None, lon_o=None, lat_o=None,
        degree=2, epsg=3006,
        **kwargs):
    """
    LEVEL-2 data
    ------------
    lon0, lat0 :: center longitude/latitude of ground pixels
    field      :: satellite observations (e.g. XCO2) - without noise
    clouds     :: cloud fractions at ground pixels
    is_plume   :: boolean array where plume has been detected
    t_values   :: t-values from plume detection
    noise      :: noise on field

    MODEL DATA (e.g COSMO)
    ----------------------
    lon_model    :: longitude in model (lon, lat)
    lat_model    :: latitude in model (lon, lat)
    height_model :: height in model (lon, lat, layers+1)
    uwind        :: u-wind in model (lon, lat, layers)
    vwind        :: v-wind in model (lon, lat, layers)

    FLUX INVERSION PARAMETERS
    -------------------------
    origin       :: name of plume (used to get lon_o and lat_o)
    lon_o, lat_o :: longitude/latitude of point source
    degree       :: degree of polynomial fitted on plume

    """
    # TRANSFORM LON/LAT TO CARTESIAN COORDS
    if origin is not None:
        lon_o, lat_o = plume.get_source_origin(origin)

    i_o, j_o = misc.nearest_index(lon_o, lat_o, lon0, lat0)
    x_o, y_o = 0.0, 0.0

    x_sat, y_sat = transform_to_kilometers(lon0, lat0, lon_o, lat_o, epsg)
    x_model, y_model = transform_to_kilometers(lon_model, lat_model, lon_o, lat_o, epsg)


    # DETERMINE CENTER LINE OF PLUME 
    # get location of plume and XCO2 values (plus wind etc.)
    x_p = x_sat[is_plume]
    y_p = y_sat[is_plume]
    t_p = t_values[is_plume]

    # origin of center line
    x0, y0 = -x_p.mean(), -y_p.mean()
    print('center line origin:', x0, y0)

    x_curve, y_curve, f_curve, c_curve = fit_2D_curve(x_p, y_p, t_p,
            degree=degree, x0=x0, y0=y0)


    #return x_curve, y_curve, f_curve, c_curve, x_p, y_p


    # DISTANCE FROM CENTER LINE
    # make plume bigger to include background
    struct = ndi.generate_binary_structure(2, 2)
    full_area = ndi.binary_dilation(is_plume, struct, iterations=3) # TODO iterations?

    # TODO
    full_area &= (clouds <= 0.01)

    # compute for all ground pixels distance to center line
    x_a = x_sat[full_area]
    y_a = y_sat[full_area]
    field_a = field[full_area] + noise[full_area]

    points = np.transpose([x_a, y_a])
    dx, dy = compute_arc_length_and_distance(points, f_curve, c_curve,
            x0=x0, y0=y0)


    # remove points downstream of lon_p.max()
    x_a = x_sat[full_area][np.isfinite(dx)]
    y_a = y_sat[full_area][np.isfinite(dx)]
    field_a = field[full_area][np.isfinite(dx)] + noise[full_area][np.isfinite(dx)]

    dy = dy[np.isfinite(dx)]
    dx = dx[np.isfinite(dx)]

    ##################################
    # WIND SPEED ALONG CENTER LINE   #
    ##################################

    i_o, j_o = misc.nearest_index(x_o, y_o, x_model, y_model)

    # estimate cardinal wind direction (and plume height) 
    # along center line (TODO)
    uwind = uwind[::-1, i_o, j_o]
    vwind = vwind[::-1, i_o, j_o]
    hh = height_model[::-1, i_o, j_o]
    hh = 0.5 * (hh[1:] + hh[:-1])

    u_mean = np.mean(uwind[hh<5000.0])
    v_mean = np.mean(vwind[hh<5000.0])

    print(np.sqrt(u_mean**2 + v_mean**2))


    # wind direction profile
    dd = (270.0 - np.rad2deg(np.arctan2(vwind,uwind))) % 360

    # wind direction estimate from center line
    # TODO: slope does only knows westerly winds
    xt,yt = f_curve(c_curve, x=x_o, y=y_o, m=1, x0=x0, y0=y0)
    d_o = (270 - np.rad2deg(np.arctan2(yt,xt))) % 360
    print(d_o)


    # estimate wind direction (at center line) TODO
    heights = find_all_roots(hh, dd, d_o, h_max=5000.0)
    if len(heights) != 1:
        print('Warning: found %d heights with fitting wind direction'
                % len(heights))
        print(heights)

    if heights:
        h_o = heights[0]

        # compute wind speed at origin (TODO: wind speed along center line)
        u_o = np.interp(h_o, hh, uwind)
        v_o = np.interp(h_o, hh, vwind)
        wind = np.sqrt(u_o**2 + v_o**2)

    else:
        wind = np.sqrt(u_mean**2 + v_mean**2)

    return x_curve, y_curve, x_a, y_a, field_a, dx, dy, wind, \
            f_curve, c_curve


    # TODO: distance and wind speed can be used to determine age of plume













