
from six import string_types
from datetime import datetime, timedelta
import itertools
import os

from PIL import Image
from skimage import draw
from skimage import measure

import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage as ndi
import scipy.stats as stats
import scipy.optimize as opt
import pandas

from . import io
from . import map
from . import misc
from . import config as cfg

# location of Berlin
BERLIN = 13.407696, 52.516984


# ignore invalid floating-point operation
np.seterr(invalid='ignore')




def polyfit_on_plume(is_plume, degree=2, io=None, jo=None):

    n,m = is_plume.shape
    i,j = np.meshgrid(np.arange(m), np.arange(n))

    i = i[is_plume]
    j = j[is_plume]
    w = np.ones(i.size)

    if io is not None and jo is not None:
        i = np.append(i, io)
        j = np.append(j, jo)
        w = np.append(w, 100.0)

    return i,j

    #TODO
    t = np.arange(i.size)

    # x(t)   
    c = np.polyfit(t,i, degree, w=w)
    x = np.polyval(c,t)

    # y(t)   
    c = np.polyfit(t,j, degree, w=w)
    y = np.polyval(c,t)

    return x,y



def find_plume_by_signal(signal, thr, lon0=BERLIN[0], lat0=BERLIN[1],
        lon=None, lat=None, dn=10, clouds=None, cloud_thr=0.01):
    """\
    Find plume that has signal above threshold and is connected to location
    given by lon0 and lat0.
    """
    if clouds is None:
        is_plume = signal > thr
    else:
        is_plume = (signal > thr) & (clouds < cloud_thr)

    labels = measure.label(is_plume, background=0)
    return find_plume_at_location(labels, lon0, lat0, lon, lat, dn)



def find_plume_at_location(labels, lon_o, lat_o, lon, lat, dn=10):

    try:
        s = get_area_around_source(lon_o, lat_o, lon, lat, dn)
    except ValueError:
        return np.zeros(labels.shape, bool)

    numbers = set(labels[s].flatten()) - {0}

    if numbers:
        return  np.any([labels == i for i in numbers], axis=0)
    else:
        return np.zeros(labels.shape, bool)


def confusion_matrix(detection, ground_truth):

    TP = np.sum(detection & ground_truth)
    TN = np.sum(~detection & ~ground_truth)
    FP = np.sum(detection & ~ground_truth)
    FN = np.sum(~detection & ground_truth)

    return TP, FP, FN, TN



def label_plumes(d, n_min=20):
    """
    Label detected plume pixels.
    """
    d[np.isnan(d)] = 0
    labels = measure.label(d, background=0)

    i = 1
    final = np.zeros_like(labels)

    for l in set(labels.flatten()):
        if l != 0 and np.sum(labels == l) >= n_min:
            final[labels == l] = i
            i += 1

    return final


def welch_satterthwaite_equation(s1, s2, n1, n2):

    a = (s1**2 / n1 + s2**2 / n2)**2
    b = s1**4 / (n1**2 * (n1-1)) + s2**2 / (n2**2 * (n2-1))

    return a/b


def local_stats(img, shape=(3,3)):

    def nanmean(x):
        n = np.sum(np.isfinite(x))
        return np.nanmean(x) if n>0 else np.nan

    def nanvar(x):
        n = np.sum(np.isfinite(x))
        return np.nanvar(x, ddof=1) if n>1 else np.nan


    domain = np.any(img, axis=1)

    mean = np.full(img.shape, np.nan)
    mean[domain,:] = ndi.generic_filter(
            img[domain,:], nanmean, shape, mode='constant', cval=np.nan
    )
    var = np.full(img.shape, np.nan)
    var[domain,:] = ndi.generic_filter(
            img[domain,:], nanvar, shape, mode='constant', cval=np.nan
    )

    n = np.full(img.shape, np.nan)
    n[domain,:] = ndi.generic_filter(
            np.isfinite(img).astype(int)[domain,:], np.count_nonzero, shape,
            mode='constant', cval=0
    )

    return mean, var, n





def get_area_around_source(lon_o, lat_o, lon, lat, bg_area):
    # define area of background around Berlin

    i,j = misc.nearest_index(lon_o, lat_o, lon, lat)

    d = (np.sqrt((lon[i,j] - lon_o)**2 + (lat[i,j] - lat_o)**2))

    if d > 0.1: # about 10 km
        raise ValueError('source (likely) outside swath')

    imin, imax = max(0, i-bg_area), i+bg_area
    jmin, jmax = max(0, j-bg_area), j+bg_area
    bg = slice(imin, imax), slice(jmin, jmax)

    return bg


def detect(lon, lat, field, is_cloudy, footprint,
        lon_o, lat_o, source_area=1,
        p_thr=0.997, sigma=None,
        bg_values=None, bg_area=50,
    ):
    """\
    Detect plume in `field` at location (`lon_o`, `lat_o`)

    Parameters
    ''''''''''
    lon :: np.ndarray
        longitudes at ground pixel centers

    lat :: np.ndarray
        latitudes at ground pixel centers

    field :: np.ndarray
        synthetic satellite observations (with noise)

    is_cloudy :: np.ndarray (boolean)
        array with True where clouds

    footprint :: int
        size of footprint used for computing local stats (n x n)

    lon_o, lat_o
        longitude and latitude of source origin

    source_area
        size of source area (in pixels)

    p_thr :: float
        detection threshold 

    sigma :: float
        single sounding precision (only used if footprint == 1 and thus
        sample variance cannot be computed from neighborhood)

    bg_values :: np.array or None
        field used for computing mean and variance of background
        if None use field

    bg_area:: int
        size of background area around origin (lon_o, lat_o)

    """

    # if footprint is even, center pixel is not defined clearly.
    if footprint % 2 == 0:
        print('Warning: footprint should be odd number (1,3,5,...).')

    # estimate background mean and variance
    bg_slice = get_area_around_source(lon_o, lat_o, lon, lat, bg_area=bg_area)

    if bg_values is None:
        bg_values = field.copy()

    bg_values = bg_values[bg_slice]

    mean_b = np.nanmean(bg_values)
    var_b = np.nanvar(bg_values)
    n_b = np.isfinite(bg_values).sum()


    # estimate local mean and variance
    if footprint == 1:
        mean_s = field.copy()
        var_s = np.full(field.shape, sigma**2)
        n_s = np.ones_like(field)
    else:
        mean_s, var_s, n_s = local_stats(field, (footprint,footprint))


    # Welch's t-test
    t_values = (mean_s - mean_b) / np.sqrt(var_s / n_s + var_b / n_b)

    # degree of freedom
    if footprint == 1:
        dof = np.ones_like(field)
    else:
        dof = welch_satterthwaite_equation(var_s, var_b, n_s, n_b)


    # t-value depends on degree of freedom (Welch's t-test)
    is_hit = find_hits(t_values, p_thr, dof)

    # label plumes
    labels = label_plumes(is_hit, 10) # TODO: de-hardcode

    is_plume = find_plume_at_location(labels, lon_o, lat_o, lon, lat, dn=source_area)

    # some pixels in `in_plume` might be detected (as neighborhood) but cloudy
    # these pixels will be set to false, because they are not part of ground
    # truth and cannot be used for flux inversion (however, they are used to
    # label regions)
    is_plume[is_plume & is_cloudy] = False


    info = {'mean': mean_s, 'var': var_s, 'number': n_s,
            'mean_bg': mean_b, 'var_bg': var_b, 'n_bg': n_b,
            'dof': dof,
            't_values': t_values, 'is_hit': is_hit,
            'labels': labels,
    }

    return is_plume, info


def get_source_origin(name):
    if name == 'Berlin':
        lon_o, lat_o = cfg.CITIES[name]
    else:
        lon_o, lat_o = cfg.POINT_SOURCES[name]

    return lon_o, lat_o


def get_signal_name(tracer, origin):

    suffix = 'BV' if origin == 'Berlin' else 'JV'

    if tracer == 'CO2':
        tracer = 'XCO2'

    return '%s_%s' % (tracer, suffix)




def plume_detection_algorithm(filename, tracer='CO2', error_std=0.7,
        cloud_thr=0.01, seed=None, use_epf=False, p_thr=0.99, footprint=3,
        bg_area=50, bg_type='model',
        origin='Berlin'):
    """\
    Origin is Berlin or others (i.e. "Janschwalde" tracer)
    """

    # lon/lat of source origin
    lon_o, lat_o = get_source_origin(origin)

    # time and groud pixels
    time = io.filename2time(filename)
    lon0, lat0, lonc, latc = io.read_ground_pixels(filename)

    # read signal, clouds and total tracer field
    signal_name = get_signal_name(tracer, origin)
    signal, clouds = io.read_fields(filename, [signal_name, 'CLCT'], time,
            correct_berlin=True)

    true_field = io.read_total_tracer(filename, tracer, time, correct_berlin=True)

    # read noise
    if seed is not None:
        np.random.seed(seed)

    noise = io.read_tracer_noise(filename, tracer, use_epf, error_std)

    # field is an array without mask (masked values should be set to nan)
    field = np.array(true_field + noise)
    field[true_field.mask | (clouds > cloud_thr)] = np.nan

    # background values
    if bg_type == 'field':
        bg_values = field.copy()
    elif bg_type == 'model':
        bg_values = field - signal
    else:
        raise ValueError("bg_type has to be 'field' or 'model' not '%s'" % bg_type)


    # source area
    if origin == 'Berlin':
        source_area = 10 # (20x20 km2)
    elif origin == 'Lippendorf':
        source_area = 5
    else:
        source_area = 2  # 4x4 km2)

    if tracer == 'NO2':
        if error_std == 'low': # low noise scenario
            error_std = 1e15
        elif error_std == 'high': # high noise scenario
            error_std = 2e15

    # detect plume
    is_plume, info = detect(
            lon0, lat0, field, (clouds > cloud_thr), footprint,
            lon_o, lat_o, source_area,
            p_thr, error_std,
            bg_values, bg_area,
    )

    info.update({
        'lon0': lon0, 'lat0': lat0,
        'lonc': lonc, 'latc': latc,
        'signal': signal, 'field': true_field,
        'noise':  noise,  'clouds': clouds,
        'lon_o': lon_o, 'lat_o': lat_o
    })

    return is_plume, info





def find_hits(t_values, p_thr, dof):
    t_thr = stats.t.ppf(p_thr, dof)
    is_hit = t_values > t_thr
    return is_hit


def visualize(tracer, lonc, latc, signal, field, noise, clouds, is_plume, labels,
        is_berlin=None, show_clouds=True, do_zoom=True, lon0=None, lat0=None,
        use_noise=True, cloud_thr=None, marker=',', **kwargs):

    # tracer field
    if not use_noise:
        noise = 0.0

    if not show_clouds:
        clouds = None

    if clouds is not None and cloud_thr is not None:
        clouds = clouds > cloud_thr

    if tracer == 'CO2':
        fig = map.make_level2_map(
            lonc, latc, field - field.mean() + noise, vmin=-2, vmax=2,
            label='XCO$_2$ signal (ppm)', clct=clouds,
        )
    else:
        fig = map.make_level2_map(
            lonc, latc, field / 1e15, vmin=0, vmax=20,
            label='NO$_2$ column (10$^{15}$ molecules per cm$^2$)', clct=clouds,
        )
    #
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    if is_plume is not None and np.sum(is_plume) > 0:
        ms = 3
        ax.plot(lon0[is_plume], lat0[is_plume], 'k', marker=marker, ls='', ms=ms,
            transform=map.ccrs.PlateCarree()
    )

    if do_zoom:
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)


def show_all_ground_truths(filename, thr=0.1, cloud_thr=0.01, dn=10,
        output_path=None):

    # time and groud pixels
    time = io.filename2time(filename)
    lon, lat, lonc, latc = io.read_ground_pixels(filename)

    bv,jv,clct = io.read_fields(filename, ['XCO2_BV', 'XCO2_JV', 'CLCT'])

    names = ['Janschwalde', 'Schwarze Pumpe', 'Boxberg',
            'Lippendorf', 'Turow', 'Patnow']
    colors = [
        'grey',
        '#1f77b4', # Berlin
        '#ff7f0e', # J
        '#2ca02c', # SP
        '#d62728', # Bb
        '#9467bd', # Ld
        '#8c564b', # T
        '#e377c2', # P
        #'#7f7f7f',
        '#bcbd22',
        #'#17becf',
    ]
    labels = [
        'back-\nground',
        'Berlin',
        'Jänsch-\nwalde',
        'Schwarze\nPumpe',
        'Boxberg',
        'Lippen-\ndorf',
        'Turów',
        'Patnów',
        'ambig-\nuous'
    ]

    # init mask
    mask = np.zeros(lon.shape, dtype=int)

    # Berlin plume
    lon_o, lat_o = cfg.CITIES['Berlin']
    p = find_plume_by_signal(bv, thr, lon_o, lat_o, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)
    mask[p] = 1

    # power plant plumes
    shape = mask.shape + (6,)
    plants = np.zeros(shape, dtype=int)

    for i, name in enumerate(names, 2):
        lon_o, lat_o = cfg.POINT_SOURCES[name]
        p = find_plume_by_signal(jv, thr, lon_o, lat_o, lon, lat, dn=dn,
                            clouds=clct, cloud_thr=cloud_thr)

        plants[p,i-2] = i
        mask[p] = i

    # ambiguous plunmes
    mask[np.sum(plants > 0, axis=2) > 1] = len(names)+2

    labels2 = [labels[0]]
    for i, label in enumerate(labels[1:], 1):
        label = '%s\n[%d]' % (label, np.sum(mask == i))
        labels2.append(label)
    labels = labels2

    # nothing detected
    if not np.any((mask != 0) & (clct < cloud_thr)):
        return None

    # reverse numbering, colors and labels
    mask = -1 * mask + len(names) + 2
    colors = colors[::-1]
    labels = labels[::-1]

    # plot
    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=clct, is_discrete=True, cmap=colors,
        cb_labels=labels
    )

    ax, cax = fig.get_axes()
    map.add_city_labels(ax, size='medium')

    text = io.filename2label(filename)
    ax.text(0.0, 1.0, text, ha='left', va='top',
             bbox=dict(facecolor='white', alpha=1.0),
             transform=ax.transAxes
    )

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


    if output_path is not None:
        filename = os.path.basename(filename)
        filename = filename.replace('.nc', '_plumes.png')
        filename = os.path.join(output_path, filename)

        plt.savefig(filename)

    return True



def show_ground_truth(lonc, latc, signal, clouds, do_zoom=True, show_clouds=True,
        thr=0.1, cloud_thr=0.01, mask_clouds=True, lon_o=BERLIN[0], lat_o=BERLIN[1],
        dn=10, **kwargs):

    if mask_clouds:
        mask = find_plume_by_signal(signal, thr, lon_o, lat_o,
                lonc.mean(2), latc.mean(2), dn=dn,
                clouds=clouds, cloud_thr=cloud_thr)
    else:
        mask = find_plume_by_signal(signal, thr, lon_o, lat_o,
                lonc.mean(2), latc.mean(2), dn=dn)

    n_pixels = np.sum(mask)

    mask = (~mask).astype(int)
    if show_clouds:
        mask[clouds > cloud_thr] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume\n[%d]' % n_pixels, 'back-\nground', 'clouds']
    )
    ax, cax = fig.get_axes()
    map.add_city_labels(ax)

    if do_zoom:
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)

    text = 'XCO$_2^\mathrm{thr}$ = %.2f ppm, d$_\mathrm{max}$ = %d px' % (thr, dn)
    ax.text(0.01, 0.0, text, ha='left', va='bottom', fontsize='x-small',
             transform=ax.transAxes
    )


def show_labels(lonc, latc, labels, clouds, do_zoom=True, **kwargs):
    # labels
    fig = map.make_level2_map(
        lonc, latc, labels, vmin=None, vmax=None,
        label='labels of detected plumes', clct=clouds > 0.01, is_discrete=True,
        cmap='nipy_spectral_r'
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)


def show_plume(lonc, latc, is_plume, clouds, do_zoom=True, **kwargs):

    mask = np.ones(is_plume.shape, int)
    mask[is_plume] = 0
    mask[clouds > 0.01] = 2

    fig = map.make_level2_map(
        lonc, latc, mask, vmin=None, vmax=None,
        label='', clct=None, is_discrete=True, cmap='Greys_r',
        cb_labels=['plume', 'back-\nground', 'clouds'],
    )
    if do_zoom:
        ax, cax = fig.get_axes()
        ax.set_xlim(1, 3.8)
        ax.set_ylim(3.5, 6)
        map.add_city_labels(ax)


def plot_tp(filenames, n=4, p_thr=0.997):

    m = []
    totals = []
    labels = []

    for filename in filenames:

        sat, orbit, date, lon = io.process_filename(filename)

        label = '%s, #%d' % (date.strftime('%d %b'), lon//90 + 1)
        labels.append(label)

        d = pandas.read_csv(filename, sep='\t')
        d = d[np.isclose(d['p_thr'], p_thr) & (d['neighborhood'] == n)]

        N = np.mean(d['TP'] + d['FN'])

        m.append( d['TP'] )
        totals.append(N)

    return labels, np.array(m), np.array(totals, dtype=int)




def plot_performance(filename, p_thr=0.997):
    """\
    Plot performance for a single file.
    """
    d = pandas.read_csv(filename, sep='\t')
    d = d[np.isclose(d['p_thr'], p_thr)]

    n = np.array(sorted(set(d['neighborhood'])))

    dn = [d[d['neighborhood'] == i]  for i in n]

    N = np.mean(d['TP'] + d['FN'])

    # true positive rate
    TPR = [v['TP'] / (v['TP'] + v['FN']) for v in dn]

    plt.figure()

    plt.boxplot(TPR)
    plt.ylim(0,1.1)
    plt.grid()

    plt.xlabel(r'size of neighborhood (n$\times$n)')
    plt.ylabel('true positive rate')

    #plt.boxplot(FDR)
    #plt.ylim(0,1.1)

    plt.twinx()
    plt.ylim(0,1.1*N)
    plt.ylabel('true positive')


    plt.figure()

    # false discovery rate
    FDR = [v['FP'] /(v['TP'] + v['FP']) for v in dn]
    plt.boxplot(FDR)
    plt.ylabel('false discovery rate')




def main(input_fname, output_fname, tracer, error_type, use_epf=False, cloud_thr=0.01,
        bg_type='model', lon_o=BERLIN[0], lat_o=BERLIN[1]):

    # CO2 (error_type: 0.5, 0.7, 1.0, use_epf: True or False)
    if tracer == 'CO2':
        signal_thr = 0.1

    # NO2 (error: low or high, cloud_thr: 0.01 or 0.30, use_epf: True or False
    elif tracer == 'NO2':
        signal_thr = 0.2e15

    # CO
    elif tracer == 'CO':
        signal_thr = 0.5e17

    # used for all tracers
    p_thr     = [0.95, 0.997]
    n         = [2,3,4,5,6,7,8,9,10]
    seed = None
    plume_name = 'Berlin'

    headings  = [
            'filename',     # Level-2 filename
            'plume_name',   # name of plume (Berlin, Janschwalde, ...)
            'tracer',       # tracer (XCO2, NO2 or CO)
            'uncertainty',  # base uncertainty of tracer
            'use_epf',      # use XCO2 or NO2 error formula (Buchwitz or Wenig)
            'tracer_thr',   # threshold for defining ground truth
            'cloud_thr',    # threshold for masking clouds
            'neighborhood', # size of neighborhood (n x n)
            'p_thr',        # threshold for detection probability (one-side t-test)
            'bg_type',      # how the background mean and var is computed
            'number',       # ensemble number
            'TP',           # true positives
            'FP',           # false positives
            'FN',           # false negaitves
            'TN',           # true negatives
    ]
    with open(output_fname, 'w') as out:
        out.write('\t'.join(headings))
        out.write('\n')

        # loop over parameters
        for p_thr, n in itertools.product(p_thr, n):
            for i in range(10):

                is_plume, info =  plume_detection_algorithm(input_fname, tracer, error_type,
                        cloud_thr, seed, use_epf, p_thr, n,
                        bg_type=bg_type, bg_area=50)


                lon = info['lon0']
                lat = info['lat0']
                signal = info['signal']
                clouds = info['clouds']

                # get ground truth
                if i == 0:
                    ground_truth = find_plume_by_signal(signal, signal_thr,
                            lon_o, lat_o, lon, lat, dn=10,
                            clouds=clouds, cloud_thr=cloud_thr)

                # compute performance
                TP, FP, FN, TN = confusion_matrix(is_plume, ground_truth)

                # save results
                line = [
                    os.path.basename(input_fname),
                    plume_name,
                    tracer,
                    error_type,
                    use_epf,
                    signal_thr,
                    cloud_thr,
                    n,
                    p_thr,
                    bg_type,
                    i,
                    TP,
                    FP,
                    FN,
                    TN
                ]
                #print('\t'.join(str(v) for v in line))
                out.write('\t'.join(str(v) for v in line))
                out.write('\n')

            out.flush()


if __name__ == '__name__':
    pass

